/* tag: forth bootstrap environment
 *
 * Copyright (C) 2003, 2004 Patrick Mauritz, Stefan Reinauer
 *
 * See the file "COPYING" for further information about
 * the copyright and warranty status of this work.
 */

#include "openbios/sysinclude.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <sys/stat.h>

#ifdef __GLIBC__
#define _GNU_SOURCE
#include <getopt.h>
#endif

#include "openbios/config.h"
#include "openbios/stack.h"
#include "openbios/sysinclude.h"
#include "openbios/kernel.h"
#include "dict.h"

#define MEMORY_SIZE (1024*1024)	/* 1M ram for hosted system */
#define DICTIONARY_SIZE (256*1024) /* 256k for the dictionary   */

/* local variables */

ucell *latest, *state;
ucell *memory;

static int errors = 0;
static int segfault = 0;
int verbose = 0;

static FILE *srcfiles[64];
static int cursrc = -1;

/* include path handling */
typedef struct include_path include;
struct include_path {
	char *path;
	include *next;
};

include includes = { ".", NULL };

ucell * relocation_address=NULL;
int     relocation_length=0;
/* the word names are used to generate the prim words in the
 * dictionary. This is done by the C written interpreter.
 */

static const char *wordnames[] = {
	"(semis)", "", "(lit)", "", "", "", "", "(do)", "(?do)", "(loop)",
	"(+loop)", "", "", "", "dup", "2dup", "?dup", "over", "2over", "pick", "drop",
	"2drop", "nip", "roll", "rot", "-rot", "swap", "2swap", ">r", "r>",
	"r@", "depth", "depth!", "rdepth", "rdepth!", "+", "-", "*", "u*",
	"mu/mod", "abs", "negate", "max", "min", "lshift", "rshift", ">>a",
	"and", "or", "xor", "invert", "d+", "d-", "m*", "um*", "@", "c@",
	"w@", "l@", "!", "+!", "c!", "w!", "l!", "=", ">", "<", "u>", "u<",
	"sp@", "move", "fill", "(emit)", "(key?)", "(key)", "execute",
	"here", "here!", "dobranch", "do?branch", "unaligned-w@",
	"unaligned-w!", "unaligned-l@", "unaligned-l!", "ioc@", "iow@",
	"iol@", "ioc!", "iow!", "iol!", "i", "j", "call", "sys-debug",
	"$include", "$encode-file"
};

/*
 * skipws skips all whitespaces (space, tab, newline) from the input file
 */

static void skipws(FILE * f)
{
	int c;
	while (!feof(f)) {
		c = getc(f);

		if (c == ' ' || c == '\t' || c == '\n')
			continue;

		ungetc(c, f);
		break;
	}
}

/*
 * parse gets the next word from the input stream, delimited by
 * delim. If delim is 0, any word delimiter will end the stream
 * word delimiters are space, tab and newline. The resulting word
 * will be put zero delimited to the char array line.
 */

static int parse(FILE * f, char *line, char delim)
{
	int cnt = 0, c;

	while (!feof(f)) {
		c = getc(f);

		if (delim && c == delim)
			break;

		if ((!delim) && (c == ' ' || c == '\n' || c == '\t'))
			break;

		line[cnt++] = c;
	}

	line[cnt] = 0;

	return cnt;
}

/*
 * parse_word is a small helper that skips whitespaces before a word.
 * it's behaviour is similar to the forth version parse-word. 
 */

static void parse_word(FILE * f, char *line)
{
	skipws(f);
	parse(f, line, 0);
}


static void writestring(const char *str)
{
	unsigned int i;
	for (i = 0; i < strlen(str); i++) {
		dict[dicthead + i] = str[i];
	}
	dicthead += i + 1;
	dict[dicthead - 1] = (char) strlen(str) + 128;
}

#define writebyte(value) {*(dict+dicthead++)=value;}
#define writecell(value) {*(cell *)(dict+dicthead)=value; dicthead+=sizeof(cell);}
static void reveal(void)
{
	*last = *latest;
}


static void paddict(ucell align)
{
	while (dicthead % align != 0)
		writebyte(0);
}

static void relocation_table(char * dict_one, char *dict_two, int length)
{
	ucell *d1=(ucell *)dict_one, *d2=(ucell *)dict_two;
	ucell *reloc_table;
	int pos, bit;
	int l=(length+(sizeof(cell)-1))/sizeof(ucell), i;
	
	/* prepare relocation table */
	relocation_length=(l+BITS-1)/BITS;
	reloc_table = malloc(relocation_length*sizeof(cell));
	memset(reloc_table,0,relocation_length*sizeof(cell));
	
	for (i=0; i<=l; i++) {
		
		pos=i/BITS;
		bit=i&~(-BITS);
		
		if(d1[i]==d2[i]) {
			reloc_table[pos] &= ~(1UL<<bit);
			
			if(d1[i] >= (ucell)dict_one &&
					d1[i] <= (ucell)(dict_one+length))
				printk("\nWARNING: wrong reloc!!");
		} else {
			/* This is a pointer, it needs relocation, d2==dict */
			reloc_table[pos] |= (1UL<<bit);
			d2[i] -= (ucell)d2;
		}
	}
#ifdef CONFIG_DEBUG_DICTIONARY
	printk("dict1 %lx dict2 %lx dict %lx\n",dict_one, dict_two, dict);
	for (i=0; i< relocation_length ; i++)
		printk("reloc %d %lx\n",i+1, reloc_table[i]);
#endif
	relocation_address=reloc_table;
}

static void write_dictionary(char *filename)
{
	FILE *f;

	dictionary_header_t header = {
		.signature	= DICTID,
		.version	= 2,
		.cellsize	= sizeof(cell),
#ifdef CONFIG_LITTLE_ENDIAN
		.endianess	= 0,
#endif
#ifdef CONFIG_BIG_ENDIAN
		.endianess	= -1,
#endif
		.compression	= 0,
		.relocation	= -1,
		.length		= dicthead,
		.last		= (ucell)last-(ucell)dict,
	};
	
	// FIXME Calculate Checksum
	
	f = fopen(filename, "w");
	if (!f) {
		printk("panic: can't write to dictionary '%s'.\n", filename);
		exit(1);
	}
	if (!relocation_address)
		header.relocation=0;
	
	dump_header(&header);

	fwrite(&header, sizeof(dictionary_header_t), 1, f);
	fwrite(dict, dicthead, 1, f);
#ifdef CONFIG_DEBUG_DICTIONARY
	printk("writing %d reloc cells \n",relocation_length);
#endif
	if(relocation_address) {
		fwrite(relocation_address, relocation_length*sizeof(cell), 1, f);
		free(relocation_address);
		relocation_address=NULL;
	}

	fclose(f);

#ifdef CONFIG_DEBUG_DICTIONARY
	printk("wrote dictionary to file %s.\n", filename);
#endif
}

static ucell read_dictionary(char *fil)
{
	int ilen;
	ucell ret;
	char *mem;
	FILE *f;
	struct stat finfo;

	if (stat(fil, &finfo))
		return 0;

	ilen = finfo.st_size;

	if ((mem = malloc(ilen)) == NULL) {
		printk("panic: not enough memory.\n");
		exit(1);
	}

	f = fopen(fil, "r");
	if (!f) {
		printk("panic: can't open dictionary.\n");
		exit(1);
	}

	if (fread(mem, ilen, 1, f) != 1) {
		printk("panic: can't read dictionary.\n");
		fclose(f);
		exit(1);
	}
	fclose(f);

	ret = load_dictionary(mem, ilen);

	free(mem);
	return ret;
}

static void fcreate(char *word, ucell cfaval)
{
	if (strlen(word) != 0) {
		writestring(word);
		/* get us at least 1 byte for flags */
		writebyte(0);
		paddict(sizeof(cell));
		/* set flags high bit. */
		dict[dicthead - 1] = 128;
		/* lfa and cfa */
		writecell(*latest);
		*latest = (ucell) dict + dicthead - sizeof(cell);
		writecell(cfaval);
	}
}


static ucell *buildvariable(char *name, cell defval)
{
	fcreate(name, DOVAR);	/* see dict.h for DOVAR and other CFA ids */
	writecell(defval);
	return (ucell *) (dict + dicthead - sizeof(cell));
}

static void buildconstant(char *name, cell defval)
{
	fcreate(name, DOCON);	/* see dict.h for DOCON and other CFA ids */
	writecell(defval);
}

static void builddefer(char *name)
{
	fcreate(name, DODFR);	/* see dict.h for DODFR and other CFA ids */
	writecell(0);
	writecell((cell)findword("(semis)"));
}

static cell getvariable(char *name)
{
	xt_t cfa = findword(name);
#ifdef CONFIG_DEBUG_INTERPRETER
	printk("getting variable %s, cfa=%p\n", name, cfa);
#endif
	return *((cell*)cfa + 1);
}

static void add_includepath(char *path)
{
	include *incl = &includes;
	include *newpath;

	while (incl->next)
		incl = incl->next;

	newpath = malloc(sizeof(include));
	if (!newpath) {
		printk("panic: not enough memory for include path.\n");
		exit(1);
	}

	incl->next = newpath;
	newpath->path = path;
	newpath->next = NULL;
}

static FILE *fopen_include(const char *fil)
{
#define MAX_PATH_LEN 256
	char fullpath[MAX_PATH_LEN];
	FILE *ret;
	include *incl = &includes;

	while (incl) {
		strncpy(fullpath, incl->path, MAX_PATH_LEN);
		strncat(fullpath, "/", MAX_PATH_LEN);
		strncat(fullpath, fil, MAX_PATH_LEN);
		ret = fopen(fullpath, "r");
		if (ret != NULL)
			return ret;
		incl = incl->next;
	}
	return NULL;
}

static int interpret_source(char *fil)
{
	FILE *f;
	char tib[160];
	long num;
	char *test;

	const cell SEMIS = (cell)findword("(semis)");
	const cell LIT = (cell)findword("(lit)");
	const cell DOBRANCH = (cell)findword("dobranch");

	if ((f = fopen_include(fil)) == NULL) {
		printk("error while loading source file '%s'\n", fil);
		errors++;
		exit(1);
	}

	/* FIXME: We should read this file at 
	 * once. No need to get it char by char 
	 */

	while (!feof(f)) {
		xt_t res;

		parse_word(f, tib);

		/* if there is actually no word, we continue right away */
		if (strlen(tib) == 0) {
			continue;
		}

		/* Checking for builtin words that are needed to
		 * bootstrap the forth base dictionary.
		 */

		if (!strcmp(tib, "(")) {
			parse(f, tib, ')');
			continue;
		}

		if (!strcmp(tib, "\\")) {
			parse(f, tib, '\n');
			continue;
		}

		if (!strcmp(tib, ":")) {
			parse_word(f, tib);

#ifdef CONFIG_DEBUG_INTERPRETER
			printk("create colon word %s\n\n", tib);
#endif
			fcreate(tib, DOCOL);	/* see dict.h for DOCOL and other CFA ids */
			*state = (ucell) (-1);
			continue;
		}

		if (!strcmp(tib, ";")) {
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("finish colon definition\n\n");
#endif
			writecell((cell)SEMIS);
			*state = (ucell) 0;
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "variable")) {
			parse_word(f, tib);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("defining variable %s\n\n", tib);
#endif
			buildvariable(tib, 0);
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "constant")) {
			parse_word(f, tib);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("defining constant %s\n\n", tib);
#endif
			buildconstant(tib, POP());
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "value")) {
			parse_word(f, tib);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("defining value %s\n\n", tib);
#endif
			buildconstant(tib, POP());
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "defer")) {
			parse_word(f, tib);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("defining defer word %s\n\n", tib);
#endif
			builddefer(tib);
			reveal();
			continue;
		}

		if (!strcasecmp(tib, "include")) {
			parse_word(f, tib);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("including file %s\n\n", tib);
#endif
			interpret_source(tib);
			continue;
		}

		if (!strcmp(tib, "[']")) {
			xt_t xt;
			parse_word(f, tib);
			xt = findword(tib);
			if (*state == 0) {
#ifdef CONFIG_DEBUG_INTERPRETER
				printk
				    ("writing address of %s to stack\n\n",
				     tib);
#endif
				PUSH_xt(xt);
			} else {
#ifdef CONFIG_DEBUG_INTERPRETER
				printk("writing lit, addr(%s) to dict\n\n",
				       tib);
#endif
				writecell(LIT);	/* lit */
				writecell((cell)xt);
			}
			continue;
			/* we have no error detection here */
		}

		if (!strcasecmp(tib, "s\"")) {
			int cnt;
			cell loco;

			cnt = parse(f, tib, '"');
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("compiling string %s\n", tib);
#endif
			loco = dicthead + (6 * sizeof(cell));
			writecell(LIT);
			writecell((ucell) dict + loco);
			writecell(LIT);
			writecell(cnt);
			writecell(DOBRANCH);
			loco = cnt + sizeof(cell) - 1;
			loco &= ~(sizeof(cell) - 1);
			writecell(loco);
			memcpy(dict + dicthead, tib, cnt);
			dicthead += cnt;
			paddict(sizeof(cell));
			continue;
		}

		/* look if tib is in dictionary. */
		/* should the dictionary be searched before the builtins ? */
		res = findword(tib);
		if (res) {
			u8 flags = *((u8*)res - sizeof(cell) - 1);
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("%s is 0x%p\n", tib, (ucell) res);
#endif
			if (!(*state) || (flags & 3)) {
#ifdef CONFIG_DEBUG_INTERPRETER
				printk("executing %s, %p (flags: %s %s)\n",
				       tib, res,
				       (flags & 1) ? "immediate" : "",
				       (flags & 2) ? "compile-only" : "");
#endif
				PC = (ucell)res;
				enterforth(res);
			} else {
#ifdef CONFIG_DEBUG_INTERPRETER
				printk("writing %s to dict\n\n", tib);
#endif
				writecell((cell)res);
			}
			continue;
		}


		/* if not look if it's a number */
		if (tib[0] == '-')
			num = strtol(tib, &test, getvariable("base"));
		else
			num = strtoul(tib, &test, getvariable("base"));

		if (*test != 0) {
			/* what is it?? */
			printk("%s is not defined.\n\n", tib);
			errors++;
#ifdef CONFIG_DEBUG_INTERPRETER
			continue;
#else
			return -1;
#endif
		}

		if (*state == 0) {
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("pushed %x to stack\n\n", num);
#endif
			PUSH(num);
		} else {
#ifdef CONFIG_DEBUG_INTERPRETER
			printk("writing lit, %x to dict\n\n", num);
#endif
			writecell(LIT);	/* lit */
			writecell(num);
		}
	}
	fclose(f);

	return 0;
}

static int build_dictionary(void)
{
	ucell lfa = 0;
	unsigned int i;

	/* we need a temporary place for latest 
	 * outside the dictionary
	 */

	latest = &lfa;

	/* starting a new dictionary: clear dicthead */
	dicthead = 0;
	
#ifdef CONFIG_DEBUG_DICTIONARY
	printk("building dictionary, %d primitives.\nbuilt words:",
	       sizeof(wordnames) / sizeof(void *));
#endif

	for (i = 0; i < sizeof(wordnames) / sizeof(void *); i++) {
		if (strlen(wordnames[i]) != 0) {
			fcreate((char *) wordnames[i], i);
#ifdef CONFIG_DEBUG_DICTIONARY
			printk(" %s", wordnames[i]);
#endif
		}
	}
#ifdef CONFIG_DEBUG_DICTIONARY
	printk(".\n");
#endif

	/* get last/latest and state */
	state = buildvariable("state", 0);
	last = buildvariable("forth-last", 0);
	latest = buildvariable("latest", 0);

	*latest = (ucell) (latest - 2);

	buildvariable("base", 10);
	buildconstant("/c", sizeof(u8));
	buildconstant("/w", sizeof(u16));
	buildconstant("/l", sizeof(u32));
	buildconstant("/n", sizeof(void *));

	reveal();
	printk("finished building dictionary.\n");

	return 0;
}

/*
 * functions used by primitives
 */

int availchar(void)
{
	int tmp;
	if( cursrc < 0 ) {
		runforth = 0;
		/* return -1 in order to exit the loop in key() */
		return -1;
	}

	tmp = getc( srcfiles[cursrc] );
	if (tmp != EOF) {
		ungetc(tmp, srcfiles[cursrc]);
		return -1;
	}
	fclose( srcfiles[cursrc--] );

	return availchar();
}

int get_inputbyte( void )
{
	int tmp;
	
	if( cursrc < 0 ) {
		runforth = 0;
		return 0;
	}
	
	tmp = getc( srcfiles[cursrc] );
	if (tmp != EOF)
		return tmp;
	fclose( srcfiles[cursrc--] );

	return get_inputbyte();
}



u8 inb(u32 reg)
{
	return 0xff;
}

u16 inw(u32 reg)
{
	return 0xffff;
}

u32 inl(u32 reg)
{
	return 0xffffffff;
}

void outb(u32 reg, u8 val)
{
}

void outw(u32 reg, u16 val)
{
}

void outl(u32 reg, u32 val)
{
}

/*
 *  segmentation fault handler. linux specific?
 */

static void
segv_handler(int signo __attribute__ ((unused)),
	     siginfo_t * si, void *context __attribute__ ((unused)))
{
	static int count = 0;
	ucell addr = 0xdeadbeef;

	if (count) {
		printk("Died while dumping forth dictionary core.\n");
		goto out;
	}

	count++;

	if (PC >= (ucell) dict && PC <= (ucell) dict + dicthead)
		addr = *(ucell *) PC;

	printk("panic: segmentation violation at %p\n", (char *)si->si_addr);
	printk("dict=%p here=%p(dict+0x%x) pc=0x%x(dict+0x%x)\n",
	       dict, dict + dicthead, dicthead, PC, PC - (ucell) dict);
	printk("dstackcnt=%d rstackcnt=%d instruction=%x\n",
	       dstackcnt, rstackcnt, addr);

	printdstack();
	printrstack();

	printk("Writing dictionary core file\n");
	write_dictionary("forth.dict.core");

      out:
	exit(1);
}

/* 
 * allocate memory and prepare engine for memory management.
 */

static void init_memory(void)
{
	memset(memory, 0, MEMORY_SIZE);
	
	/* we push start and end of memory to the stack
	 * so that it can be used by the forth word QUIT
	 * to initialize the memory allocator
	 */

	PUSH((ucell) memory);
	PUSH((ucell) memory + MEMORY_SIZE);
}

void exception(cell no)
{
	switch (no) {
	case -19:
		printk(" undefined word.\n");
		break;
	default:
		printk("\nError %d occured.\n", no);
	}
	exit(1);
}


void
include_file( const char *name )
{
	FILE *file = fopen_include( name );
	if( !file ) {
		printk("\nFailed opening '%s'\n", name );
		exit(1);
	}
	if( cursrc+1 >= sizeof(srcfiles)/sizeof(srcfiles[0]) ) {
		printk("\nMaximal include depth reached!\n");
		exit(1);
	}
#ifdef CONFIG_DEBUG_INTERPRETER
	printk("Including '%s'\n", name );
#endif
	srcfiles[ ++cursrc ] = file;
}

void
encode_file( const char *name )
{
	FILE *file = fopen_include(name);
	int size;
	
	if( !file ) {
		printk("\nCan't open '%s'\n", name );
		exit(1);
	}
	fseek( file, 0, SEEK_END );
	size = ftell( file );
	fseek( file, 0, SEEK_SET );
	
	printk("\nEncoding %s [%d bytes]\n", name, size );
	fread( dict + dicthead, size, 1, file );
	PUSH( (unsigned long)dict + dicthead );
	PUSH( size );
	dicthead += size;
}


static void run_dictionary(const char *basedict)
{
	if(!basedict)
		return;

	read_dictionary((char *)basedict);
	PC = (ucell)findword("initialize");

	if (!PC)
		return;

	if(!srcfiles[0]) {
		cursrc = 0;
		srcfiles[0] = stdin;
	}

	dstackcnt=0;
	rstackcnt=0;

	init_memory();
#ifdef CONFIG_DEBUG_DICTIONARY
	printk("Jumping to dictionary...");
#endif

	runforth=-1;
	enterforth((xt_t)PC);
}

static void new_dictionary(const char *source)
{
	build_dictionary();
	
	interpret_source((char *)source);

	printk("interpretion finished. %d errors occured.\n",
	       errors);
}

/*
 * main loop
 */

#define BANNER	"OpenBIOS bootstrap kernel. (C) 2003, 2004 Patrick Mauritz, Stefan Reinauer\n"\
		"This software comes with absolutely no warranty. "\
		"All rights reserved.\n\n"


#define USAGE   "usage: %s [options] [dictionary file|source file]\n\n"

int main(int argc, char *argv[])
{
	struct sigaction sa;
	unsigned char *dictname = NULL;
	unsigned char *basedict = NULL;

	unsigned char *bootstrapdict[2];
	int c, cnt;

	const char *optstring = "VhsI:d:D:?";

	printk(BANNER);

	while (1) {
#ifdef __GLIBC__
		int option_index = 0;
		static struct option long_options[] = {
			{"version", 0, NULL, 'V'},
			{"verbose", 0, NULL, 'v'},
			{"help", 0, NULL, 'h'},
			{"include", 1, NULL, 'I'},
			{"source-dictionary", 1, NULL, 'd'},
			{"target-dictionary", 1, NULL, 'D'},
			{"segfault", 0, NULL, 's'},
		};

		c = getopt_long(argc, argv, optstring, long_options,
				&option_index);
#else
		c = getopt(argc, argv, optstring);
#endif
		if (c == -1)
			break;

		switch (c) {
		case 'V':
			printk("Version " VERSION "\n");
			return 0;
		case 'h':
		case '?':
			printk("Version " VERSION "\n" USAGE, argv[0]);
			return 0;
		case 'v':
			verbose = 1;
			break;
		case 's':
			segfault = 1;
			break;
		case 'I':
			printk("adding '%s' to include path\n", optarg);
			add_includepath(optarg);
			break;
		case 'd':
			if (!basedict) {
				printk("Using source dictionary '%s'\n", optarg);
				basedict = optarg;
			}
		case 'D':
			if(!dictname) {
				printk("Dumping final dictionary to '%s'\n", optarg);
				dictname = optarg;
			}
			break;
		default:
			return 1;
		}
	}

	if (argc < optind + 1) {
		printk(USAGE, argv[0]);
		return 1;
	}

	memory = malloc(MEMORY_SIZE);
	if (!memory) {
		printk("panic: not enough memory on host system.\n");
		exit(1);
	}

	if ((bootstrapdict[0] = malloc(DICTIONARY_SIZE)) == NULL ||
	  (bootstrapdict[1] = malloc(DICTIONARY_SIZE)) == NULL) {
		printk("panic: not enough memory.\n");
		return 1;
	}
	

	if (!segfault) {
		if (verbose)
			printk("Installing SIGSEGV handler...");

		sa.sa_sigaction = segv_handler;
		sigemptyset(&sa.sa_mask);
		sa.sa_flags = SA_SIGINFO | SA_NODEFER;
		sigaction(SIGSEGV, &sa, 0);

		if (verbose)
			printk("done.\n");
	}
	for (cnt=0; cnt<2; cnt++) {
		printk("Compiling dictionary %d/%d\n", cnt+1, 2);
		dict=bootstrapdict[cnt];
		if(!basedict) {
			new_dictionary(argv[optind]);
		} else {
			for (c=argc-1; c>=optind; c--)
				include_file(argv[c]);

			run_dictionary(basedict);
		}
	}

#ifndef CONFIG_DEBUG_INTERPRETER
	if (errors)
		printk("dictionary not dumped to file.\n");
	else
#endif
	{
		relocation_table( bootstrapdict[0], bootstrapdict[1], dicthead);
		write_dictionary( dictname ? dictname : 
					(unsigned char *)"bootstrap.dict");
	}
	
	free(memory);
	free(bootstrapdict[0]);
	free(bootstrapdict[1]);
	return 0;
}
